# -*- coding: utf-8 -*-
#
# (C) Copyright 2022 Karellen, Inc. (https://www.karellen.co/)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from os.path import exists, join as jp
from threading import RLock

from wheel_axle.runtime.constants import AXLE_DONE_FILE, AXLE_LOCK_FILE

__version__ = "0.0.2"

_DIST_INFO = "dist-info"

inter_thread_lock = RLock()


def finalize(pth_path):
    dist_info_dir = pth_path[:-3] + _DIST_INFO
    axle_done_path = jp(dist_info_dir, AXLE_DONE_FILE)

    # Double lock-check for performance
    if exists(axle_done_path):
        return

    lock_path = jp(dist_info_dir, AXLE_LOCK_FILE)

    # Lock in-process for thread race
    with inter_thread_lock:
        from filelock import FileLock  # Local import for speed

        # Lock inter-process for process race
        with FileLock(lock_path):
            # Double lock-check for performance
            if exists(axle_done_path):
                return

            # Get metadata
            from wheel_axle.runtime._symlinks import SymlinksInstaller
            from wheel_axle.runtime._axle import AxleFinalizer

            installers = [SymlinksInstaller, AxleFinalizer]  # AxleFinalizer is always last!
            for installer in installers:
                installer(dist_info_dir).run()

            # Always the last step
            try:
                os.unlink(pth_path)
            except OSError:
                # This will probably fail on Windows
                pass



# -*- coding: utf-8 -*-
#
# (C) Copyright 2022 Karellen, Inc. (https://www.karellen.co/)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import csv
import os
from typing import Tuple, List

from pip._internal.models.scheme import SCHEME_KEYS
from pip._internal.operations.install.wheel import (
    _fs_to_record_path,
    is_within_directory,
    chain,
    InstallationError
)

from wheel_axle.runtime._common import Installer
from wheel_axle.runtime.constants import SYMLINKS_FILE

Symlink = Tuple[str, str, bool]
Symlinks = List[Symlink]


def write_symlinks_file(symlinks_file: str, symlinks: Symlinks) -> None:
    with open(symlinks_file, "w") as f:
        writer = csv.writer(f)
        writer.writerows([symlink[0].replace(os.sep, "/"),
                          symlink[1].replace(os.sep, "/"),
                          int(symlink[2])]
                         for symlink in symlinks)


def read_symlinks_file(symlinks_file: str) -> Symlinks:
    with open(symlinks_file, "r") as f:
        reader = csv.reader(f)
        results = []
        for row in reader:
            results.append((row[0].replace("/", os.sep),
                            row[1].replace("/", os.sep),
                            bool(int(row[2]))))
        return results


class SymlinksInstaller(Installer):
    def install(self) -> None:
        symlinks_path = os.path.join(self.dist_info_dir, SYMLINKS_FILE)

        if not os.path.exists(symlinks_path):
            return

        symlinks = read_symlinks_file(symlinks_path)
        self._install_symlinks(symlinks)

    def _install_symlinks(self, symlinks: Symlinks):
        scheme_paths = {key: getattr(self.scheme, key) for key in SCHEME_KEYS}
        symlinking_paths = list(set(chain((self.lib_dir,), scheme_paths.values())))
        real_symlinking_paths = list(map(os.path.realpath, symlinking_paths))

        def assert_no_path_traversal(dest_dir_path: str, target_path: str) -> None:
            if not is_within_directory(dest_dir_path, target_path):
                message = (
                    "The distribution {!r} has a file {!r} trying to install"
                    " outside the target directory {!r}"
                )
                raise InstallationError(
                    message.format(self.dist_meta.project_name, target_path, dest_dir_path)
                )

        def assert_no_path_symlinking(symlink_path: str, symlink_target: str) -> None:
            for real_symlinking_path in real_symlinking_paths:
                if is_within_directory(real_symlinking_path, symlink_target):
                    return

            message = (
                "The distribution {!r} has a symlink {!r} trying to link to {!r}"
                " outside the allowed target directories {!r}"
            )
            raise InstallationError(
                message.format(self.dist_meta.project_name, symlink_path, symlink_target,
                               ",".join(real_symlinking_paths))
            )

        def make_root_scheme_path(dest: str, record_path: str) -> Tuple[str, str]:
            normed_path = os.path.normpath(record_path)
            dest_path = os.path.join(dest, normed_path)
            assert_no_path_traversal(dest, dest_path)
            return dest, dest_path

        def make_data_scheme_path(record_path: str) -> Tuple[str, str]:
            normed_path = os.path.normpath(record_path)
            try:
                _, scheme_key, dest_subpath = normed_path.split(os.path.sep, 2)
            except ValueError:
                message = (
                    "Unexpected file in {}: {!r}. .data directory contents"
                    " should be named like: '<scheme key>/<path>'."
                ).format(self.dist_meta.project_name, record_path)
                raise InstallationError(message)

            try:
                scheme_path = scheme_paths[scheme_key]
            except KeyError:
                valid_scheme_keys = ", ".join(sorted(scheme_paths))
                message = (
                    "Unknown scheme key used in {}: {} (for file {!r}). .data"
                    " directory contents should be in subdirectories named"
                    " with a valid scheme key ({})"
                ).format(self.dist_meta.project_name, scheme_key, record_path, valid_scheme_keys)
                raise InstallationError(message)

            dest_path = os.path.join(scheme_path, dest_subpath)
            assert_no_path_traversal(scheme_path, dest_path)
            return scheme_path, dest_path

        def is_data_scheme_path(path: str) -> bool:
            return path.split("/", 1)[0].endswith(".data")

        for symlink in symlinks:
            symlink_path, symlink_target, is_symlink_dir = symlink
            if is_data_scheme_path(symlink_path):
                # Data scheme
                scheme_path, norm_symlink_path = make_data_scheme_path(symlink_path)
            else:
                # Root scheme
                scheme_path, norm_symlink_path = make_root_scheme_path(self.lib_dir, symlink_path)

            # This has to be done in order one by one, because previously created symlinks
            # will affect normpath resolution
            real_symlink_target = os.path.realpath(
                os.path.join(os.path.dirname(norm_symlink_path), symlink_target))

            # Here we will compare real paths
            assert_no_path_symlinking(norm_symlink_path, real_symlink_target)

            # os.path.exists doesn't work with broken symlinks, returns False
            # so we check if something is link or it otherwise exists
            if os.path.islink(norm_symlink_path) or os.path.exists(norm_symlink_path):
                os.unlink(norm_symlink_path)

            # We preserve symlink target as de-normalized, while checking traversal as real
            os.symlink(symlink_target, norm_symlink_path, is_symlink_dir)
            self.record_installed(norm_symlink_path, norm_symlink_path, False)

    def record_installed(self,
                         srcfile: str,
                         destfile: str,
                         modified: bool = False
                         ) -> None:
        """Map archive RECORD paths to installation RECORD paths."""
        newpath = _fs_to_record_path(destfile, self.lib_dir)
        self.installed[newpath] = newpath
        if modified:
            self.changed.add(_fs_to_record_path(destfile))



# -*- coding: utf-8 -*-
#
# (C) Copyright 2022 Karellen, Inc. (https://www.karellen.co/)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import csv
import os
from email.message import Message
from typing import Dict, Set, List, cast, IO

from pip._internal.models.scheme import Scheme
from pip._internal.operations.install.wheel import (
    get_csv_rows_for_installed,
    csv_io_kwargs,
    _normalized_outrows,
    _fs_to_record_path
)
from pip._vendor.pkg_resources import Distribution

from wheel_axle.runtime._wheel import get_dist_meta, get_current_scheme, wheel_root_is_purelib

IO = IO


class Installer:
    def __init__(self, dist_info_dir: str):
        self.dist_info_dir = dist_info_dir

        self.installed: Dict[str, str] = {}
        self.changed: Set[str] = set()
        self.generated: List[str] = []

        self.dist_meta: Distribution = None
        self.wheel_meta: Message = None
        self.scheme: Scheme = None
        self.lib_dir: str = None

    def run(self) -> None:
        dist_info_dir = self.dist_info_dir
        self.dist_meta, self.wheel_meta = get_dist_meta(dist_info_dir)
        self.scheme = get_current_scheme(dist_info_dir, self.dist_meta.project_name, self.wheel_meta)

        if wheel_root_is_purelib(self.wheel_meta):
            self.lib_dir = self.scheme.purelib
        else:
            self.lib_dir = self.scheme.platlib

        self.install()

        self.finalize()

    def install(self) -> None:
        pass

    def finalize(self) -> None:
        self.update_records()

    def update_records(self):
        if not (self.installed or self.changed or self.generated):
            return

        record_text = self.dist_meta.get_metadata_lines("RECORD")
        record_rows = list(csv.reader(record_text))

        rows = get_csv_rows_for_installed(
            record_rows,
            installed=self.installed,
            changed=self.changed,
            generated=self.generated,
            lib_dir=self.lib_dir,
        )

        # Record details of all files installed
        record_path = os.path.join(self.dist_info_dir, "RECORD")

        with open(record_path, **csv_io_kwargs("w")) as record_file:
            # Explicitly cast to typing.IO[str] as a workaround for the mypy error:
            # "writer" has incompatible type "BinaryIO"; expected "_Writer"
            writer = csv.writer(cast("IO[str]", record_file))
            writer.writerows(_normalized_outrows(rows))

    def record_installed(self,
                         srcfile: str,
                         destfile: str,
                         modified: bool = False
                         ) -> None:
        """Map archive RECORD paths to installation RECORD paths."""
        newpath = _fs_to_record_path(destfile, self.lib_dir)
        self.installed[srcfile] = newpath
        if modified:
            self.changed.add(_fs_to_record_path(destfile))



# -*- coding: utf-8 -*-
#
# (C) Copyright 2022 Karellen, Inc. (https://www.karellen.co/)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from os.path import join as jp

from wheel_axle.runtime import AXLE_DONE_FILE
from wheel_axle.runtime._common import Installer


class AxleFinalizer(Installer):
    def install(self) -> None:
        self.axle_done_path = jp(self.dist_info_dir, AXLE_DONE_FILE)

        self.record_installed(self.axle_done_path, self.axle_done_path, False)
        self.update_records()

    def finalize(self) -> None:
        with open(self.axle_done_path, "wb"):
            pass



#!/usr/bin/env python
#   -*- coding: utf-8 -*-

from setuptools import setup
from setuptools.command.install import install as _install

class install(_install):
    def pre_install_script(self):
        pass

    def post_install_script(self):
        pass

    def run(self):
        self.pre_install_script()

        _install.run(self)

        self.post_install_script()

if __name__ == '__main__':
    setup(
        name = 'wheel-axle-runtime',
        version = '0.0.2',
        description = 'Axle Runtime is the runtime part of the Python Wheel enhancement library',
        long_description = '# Axle-Runtime - Python Wheel enhancement library\n\n[![Gitter](https://img.shields.io/gitter/room/karellen/Lobby?logo=gitter)](https://gitter.im/karellen/Lobby)\n[![Build Status](https://img.shields.io/github/workflow/status/karellen/wheel-axle-runtime/wheel-axle-runtime/master)](https://github.com/karellen/wheel-axle-runtime/actions/workflows/build.yml)\n[![Coverage Status](https://img.shields.io/coveralls/github/karellen/wheel-axle-runtime/master?logo=coveralls)](https://coveralls.io/r/karellen/wheel-axle-runtime?branch=master)\n\n[![PyBuilder Version](https://img.shields.io/pypi/v/wheel-axle-runtime?logo=pypi)](https://pypi.org/project/wheel-axle-runtime/)\n[![PyBuilder Python Versions](https://img.shields.io/pypi/pyversions/wheel-axle-runtime?logo=pypi)](https://pypi.org/project/wheel-axle-runtime/)\n[![PyBuilder Downloads Per Day](https://img.shields.io/pypi/dd/wheel-axle-runtime?logo=pypi)](https://pypi.org/project/wheel-axle-runtime/)\n[![PyBuilder Downloads Per Week](https://img.shields.io/pypi/dw/wheel-axle-runtime?logo=pypi)](https://pypi.org/project/wheel-axle-runtime/)\n[![PyBuilder Downloads Per Month](https://img.shields.io/pypi/dm/wheel-axle-runtime?logo=pypi)](https://pypi.org/project/wheel-axle-runtime/)\n\n# This is a companion project to [Wheel Axle/bdist_axle](https://github.com/karellen/wheel-axle)\n\n## Problem\n\n1. Python wheels do not support symlinks.\n2. PIP installation procedure is not locally extensible and does not allow adding post-install hooks.\n\n## Solution\n\n**WARNING: THIS IS EXPERIMENTAL BETA SOFTWARE. THERE ARE NO WARRANTIES OF ANY KIND. USE AT YOUR OWN RISK. ADDITIONAL\nINCLUDED DISCLAIMERS ALSO APPLY.**\n\nWheel Axle Runtime library utilizes a little-known trick used in `site.py`\'s `.pth` files that allows executing\narbitrary code while the site packages are being added. Thus, specially-crafted wheels can silently execute installed\ncode on Python interpreter startup, facilitating the "post-install hook" functionality.\n\n### Python Invariants\n\nThe core functionality relies on the following Python behaviors:\n\n* `site.py` [processes .pth files](https://github.com/python/cpython/blob/8b1b27f1939cc4060531d198fdb09242f247ca7c/Lib/site.py#L171)\n* `site.py` [executes .pth import lines](https://github.com/python/cpython/blob/8b1b27f1939cc4060531d198fdb09242f247ca7c/Lib/site.py#L186)\n* `.pth` file\'s line is executed with a local\n  variable [`fullname` denoting the `.pth` file path](https://github.com/python/cpython/blob/8b1b27f1939cc4060531d198fdb09242f247ca7c/Lib/site.py#L170)\n\nThese invariants have not changed for *18 years*.\n\n### Implementation\n\nOnce the distribution-specific `.pth` is executed by the Python interpreter, the Wheel Axle Runtime behaves as follows:\n\n1. The library checks whether a file `.dist-info/axle.done` exists. If it does it is the indication that the\n   post-install hook has executed successfully and nothing more is to be done, terminating all further processing.\n2. A process-wide *inter-thread* lock is acquired.\n3. An OS-wide *inter-process file* lock is acquired on a file `.dist-info/axle.lck`.\n4. Once the locks are acquired the `.dist-info/axle.done` existence is rechecked (double-checked locking optimization).\n5. Now that in-process and inter-process race conditions are excluded the post-install work can begin.\n6. Registered `installers` are run in sequence. Installers *should be* idempotent. The following installers are\n   currently implemented:\n    1. *Symlinks installer* processes `.dist-info/symlinks.txt`, if any.\n        1. Based on the location of the `.pth` file being executed the current installation `schema` and its paths are\n           determined. Currently, installation into a virtual environment or user location is supported and tested.\n        2. For each symlink the target path is resolved and `realpath` is used to determine the final target path.\n        3. If the symlink path and symlink target path are within one of the permitted schema locations the symlink is\n           created. Otherwise, an exception is raised and the processing is aborted.\n        4. After all symlinks are created, the `.dist-info/RECORD` file is updated to reflect the created symlinks.\n    2. *Axle installer* finalizes the installation. This installer is always executed last.\n        1. The `.dist-info/RECORD` is updated with `.dist-info/axle.done` file record.\n        2. `.dist-info/axle.done` is created.\n        3. `<distribution name and version>.pth` is then removed. If the file cannot be removed it is left in place.\n           This can happen on Windows, since the `.pth` file in question is likely opened for exclusive reading on\n           Windows.\n7. Any failure anywhere in the above process will result in an abort, an error message, and a retry the next time\n   the `.pth` will be activated.\n\n### Security\n\nThere are several security requirements and implications of having post-install hooks implemented this way.\n\n1. The installation requires write permissions to the distribution. This will be a problem if the package is installed\n   as `root` in locations such as `/usr` or `/usr/local`, or is otherwise not write-permitted, unless the post-install\n   hook is also ran with the sufficient privileges. This is generally acceptable as the primary use is considered to be\n   installation into virtual envs and user locations. That said, simply running `python -c pass` or any other python\n   invocation that does activate `site.py` under the required privileges will finalize post-install procedures.\n2. There is *an attempt* to ensure that that axle wheels symlinks and targets don\'t extend beyond the allowed `schema`\n   locations. *Those attempts are **superficial** and **have not been formally verified**.* For example, it may be\n   possible to escape the path validation/confinement by:\n    * hacking symlink creation order\n    * hacking symlink directory targets\n    * exploiting OS-specific `realpath` implementation idiosyncrasies (i.e. `strict` vs not, and what is considered\n      strict)\n\n### TODOs\n\n* Support schema detection for `home` and `prefix` installations.\n* Validate and verify Windows support.\n',
        long_description_content_type = 'text/markdown',
        classifiers = [
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: POSIX',
            'Operating System :: POSIX :: Linux',
            'Topic :: System :: Archiving :: Packaging',
            'Topic :: Software Development :: Build Tools',
            'Intended Audience :: Developers',
            'Development Status :: 4 - Beta'
        ],
        keywords = 'wheel packaging setuptools bdist_wheel symlink postinstall',

        author = 'Karellen, Inc.',
        author_email = 'supervisor@karellen.co',
        maintainer = 'Arcadiy Ivanov',
        maintainer_email = 'arcadiy@karellen.co',

        license = 'Apache License, Version 2.0',

        url = 'https://github.com/karellen/wheel-axle-runtime',
        project_urls = {
            'Bug Tracker': 'https://github.com/karellen/wheel-axle-runtime/issues',
            'Documentation': 'https://github.com/karellen/wheel-axle-runtime/',
            'Source Code': 'https://github.com/karellen/wheel-axle-runtime/'
        },

        scripts = [],
        packages = ['wheel_axle.runtime'],
        namespace_packages = [],
        py_modules = [],
        entry_points = {},
        data_files = [],
        package_data = {
            'wheel_axle/runtime': ['LICENSE']
        },
        install_requires = [
            'filelock',
            'pip'
        ],
        dependency_links = [],
        zip_safe = True,
        cmdclass = {'install': install},
        python_requires = '>=3.7',
        obsoletes = [],
    )



# -*- coding: utf-8 -*-
#
# (C) Copyright 2022 Karellen, Inc. (https://www.karellen.co/)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from email.message import Message
from email.parser import Parser
from os.path import dirname, join as jp
from typing import Tuple

from pip._internal.exceptions import UnsupportedWheel
from pip._internal.locations import get_scheme
from pip._internal.utils.virtualenv import virtualenv_no_global
from pip._vendor import pkg_resources
from pip._vendor.pkg_resources import Distribution

__all__ = ["get_current_scheme", "get_dist_meta", "wheel_root_is_purelib"]


def _get_dist(metadata_directory: str) -> Distribution:
    """Return a pkg_resources.Distribution for the provided
    metadata directory.
    """
    dist_dir = metadata_directory.rstrip(os.sep)

    # Build a PathMetadata object, from path to metadata. :wink:
    base_dir, dist_dir_name = os.path.split(dist_dir)
    metadata = pkg_resources.PathMetadata(base_dir, dist_dir)

    # Determine the correct Distribution object type.
    if dist_dir.endswith(".egg-info"):
        dist_cls = pkg_resources.Distribution
        dist_name = os.path.splitext(dist_dir_name)[0]
    else:
        assert dist_dir.endswith(".dist-info")
        dist_cls = pkg_resources.DistInfoDistribution
        dist_name = os.path.splitext(dist_dir_name)[0].split("-")[0]

    return dist_cls(
        base_dir,
        project_name=dist_name,
        metadata=metadata,
    )


def get_current_scheme(dist_info_dir, dist_name, wheel_meta):
    install_dir = dirname(dist_info_dir)

    scheme = get_scheme(dist_name)

    def check_correct_scheme(scheme, install_dir):
        if wheel_root_is_purelib(wheel_meta):
            scheme_dir = scheme.purelib
        else:
            scheme_dir = scheme.platlib

        if os.path.samefile(install_dir, scheme_dir):
            return True

        return False

    if check_correct_scheme(scheme, install_dir):
        return scheme
    else:
        if not virtualenv_no_global():
            scheme = get_scheme(dist_name, True)

            if check_correct_scheme(scheme, install_dir):
                return scheme

    raise RuntimeError("unable to determine current installation scheme")


def wheel_root_is_purelib(wheel_meta: Message) -> bool:
    return wheel_meta.get("Root-Is-Purelib", "").lower() == "true"


def get_dist_meta(dist_info_dir: str) -> Tuple[Distribution, Message]:
    return _get_dist(dist_info_dir), wheel_metadata(dist_info_dir)


def wheel_metadata(dist_info_dir) -> Message:
    """Return the WHEEL metadata of an extracted wheel, if possible.
    Otherwise, raise UnsupportedWheel.
    """
    path = jp(dist_info_dir, "WHEEL")
    with open(path, "rb") as f:
        wheel_contents = f.read()

    try:
        wheel_text = wheel_contents.decode()
    except UnicodeDecodeError as e:
        raise UnsupportedWheel(f"error decoding {path!r}: {e!r}")

    # FeedParser (used by Parser) does not raise any exceptions. The returned
    # message may have .defects populated, but for backwards-compatibility we
    # currently ignore them.
    return Parser().parsestr(wheel_text)
